use std::sync::Arc;

use axum::{
    body::Body,
    extract::Request,
    http::{header::AUTHORIZATION, StatusCode},
    middleware::Next,
    response::{IntoResponse, Response},
    Extension, Json,
};

use jsonwebtoken::{decode, Header};
use serde::{Deserialize, Serialize};
use serde_with::chrono::{DateTime, Utc};
use tracing::{debug, error, warn};
use ubisync_lib::{
    api::app::{
        AppCreatePotRequest, AppCreatePotResponse, AppGetDefaultPotRequest,
        AppGetDefaultPotResponse, AppRegisterRequest, AppRegisterResponse, AppSetDefaultPotRequest,
        AppSetDefaultPotResponse,
    },
    types::{AppId, PotId},
};

use crate::state::ApiState;

#[derive(Serialize, Deserialize, Clone, Debug, PartialEq)]
pub struct App {
    pub id: AppId,
    pub app_type: String,
    pub name: String,
    pub description: String,
    pub default_pot: Option<PotId>,
    pub last_access: DateTime<Utc>,
}

#[derive(Serialize, Deserialize, Debug)]
struct JWTClaims {
    sub: AppId,
}

pub(super) async fn auth(
    s: Extension<Arc<ApiState>>,
    mut request: Request<Body>,
    next: Next,
) -> Response {
    if let Some(auth_header) = request.headers().get(AUTHORIZATION) {
        if let Ok(header_string) = auth_header.to_str() {
            if header_string.starts_with("Bearer") {
                if let Ok(token) = decode::<JWTClaims>(
                    &header_string[7..],
                    s.jwt_decoding_key(),
                    s.jwt_validation(),
                ) {
                    if let Ok(true) = s.app_exists(token.claims.sub.clone()) {
                        debug!("Authentication for {:?} succeeded.", &token.claims.sub);
                        request.extensions_mut().insert(token.claims.sub);
                        return next.run(request).await;
                    }
                }
            }
        }
    }
    debug!("Authentication failed for request '{:?}'.", &request);
    StatusCode::UNAUTHORIZED.into_response()
}

pub(super) async fn register(
    s: Extension<Arc<ApiState>>,
    Json(body): Json<AppRegisterRequest>,
) -> Response {
    // Maybe ask for consent by user

    // If user wants registration, proceed
    let result = s.add_app(body.name, body.description, body.app_type);

    match result {
        Ok(id) => {
            // Build JWT, respond
            let jwt = jsonwebtoken::encode(
                &Header::default(),
                &JWTClaims { sub: id.clone() },
                &s.jwt_encoding_key(),
            );
            match jwt {
                Ok(token) => (
                    StatusCode::OK,
                    Json {
                        0: AppRegisterResponse {
                            token: token,
                            app_id: id,
                        },
                    },
                )
                    .into_response(),
                Err(e) => {
                    warn!("Failed to encode token: {:?}", e);
                    StatusCode::INTERNAL_SERVER_ERROR.into_response()
                }
            }
        }
        Err(e) => {
            error!("Failed to register new application! {:?}", e);
            StatusCode::INTERNAL_SERVER_ERROR.into_response()
        }
    }
}

pub(super) async fn set_default_pot(
    s: Extension<Arc<ApiState>>,
    app_id: Extension<AppId>,
    Json(body): Json<AppSetDefaultPotRequest>,
) -> Response {
    match s.set_app_default_pot(app_id.0, body.pot_id) {
        Ok(_) => (
            StatusCode::OK,
            Json {
                0: AppSetDefaultPotResponse,
            },
        )
            .into_response(),
        Err(e) => {
            debug!("Could not set default pot: {}", e);
            StatusCode::INTERNAL_SERVER_ERROR.into_response()
        }
    }
}

pub(super) async fn get_default_pot(
    s: Extension<Arc<ApiState>>,
    app_id: Extension<AppId>,
    Json(_): Json<AppGetDefaultPotRequest>,
) -> Response {
    match s.get_default_pot(app_id.0) {
        Ok(Some(p)) => (
            StatusCode::OK,
            Json {
                0: AppGetDefaultPotResponse { pot: p },
            },
        )
            .into_response(),
        Ok(_) => {
            warn!("Pot not found");
            StatusCode::NOT_FOUND.into_response()
        }
        Err(e) => {
            warn!("No default pot found: {}", e);
            StatusCode::NOT_FOUND.into_response()
        }
    }
}

pub(super) async fn create_pot(
    s: Extension<Arc<ApiState>>,
    app_id: Extension<AppId>,
    Json(body): Json<AppCreatePotRequest>,
) -> Response {
    let app = match s.get_app(app_id.0.clone()) {
        Ok(a) => a,
        Err(e) => {
            debug!("Failed to fetch app: {}", e);
            return StatusCode::INTERNAL_SERVER_ERROR.into_response();
        }
    };
    let inferred_app_type = match body.app_type {
        Some(t) => t,
        None => app.app_type,
    };
    match s.create_pot(app_id.0.clone(), inferred_app_type) {
        Ok(id) => {
            // If this is the first pot for this app, set it as default
            if app.default_pot.is_none() {
                match s.set_app_default_pot(app_id.0, id.clone()) {
                    Ok(_) => (
                        StatusCode::OK,
                        Json {
                            0: AppCreatePotResponse { pot_id: id },
                        },
                    )
                        .into_response(),
                    Err(e) => {
                        debug!("Failed to set default pot: {}", e);
                        StatusCode::INTERNAL_SERVER_ERROR.into_response()
                    }
                }
            }
            // Otherwise, just return
            else {
                (
                    StatusCode::OK,
                    Json {
                        0: AppCreatePotResponse { pot_id: id },
                    },
                )
                    .into_response()
            }
        }
        Err(e) => {
            debug!("Could not create pot: {}", e);
            StatusCode::INTERNAL_SERVER_ERROR.into_response()
        }
    }
}
