use anyhow::{anyhow, Error};
use bonsaidb::core::schema::{Collection, SerializedCollection};
use itertools::Itertools;
use serde::{Deserialize, Serialize};
use serde_with::chrono::{DateTime, Utc};
use ubisync_lib::types::{AppId, ElementId, Pot, PotId};

use crate::{
    api::v0::app::App,
    state::database::{as_key::AsKey, StateDB},
};

#[derive(Debug, Serialize, Deserialize, Collection, PartialEq, Clone)]
#[collection(name = "apps", views = [])]
pub(super) struct DbApp {
    #[natural_id]
    pub(super) id: AsKey<AppId>,
    pub(super) last_access: DateTime<Utc>,
    pub(super) app_type: String,
    pub(super) name: String,
    pub(super) description: String,
    pub(super) default_pot: Option<PotId>,
}

impl From<DbApp> for App {
    fn from(value: DbApp) -> Self {
        App {
            id: (*value.id).clone(),
            app_type: value.app_type,
            name: value.name,
            description: value.description,
            default_pot: value.default_pot,
            last_access: value.last_access,
        }
    }
}

impl StateDB {
    pub fn add_app(
        &self,
        id: AppId,
        name: String,
        description: String,
        app_type: String,
    ) -> anyhow::Result<()> {
        DbApp::push(
            DbApp {
                id: AsKey::new(id),
                last_access: Utc::now(),
                app_type,
                name,
                description,
                default_pot: None,
            },
            &self.db,
        )
        .map(|_| ())
        .map_err(|e| anyhow!(e))
    }

    pub fn set_default_pot(&self, id: AppId, pot: PotId) -> anyhow::Result<()> {
        DbApp::get(&AsKey::new(id), &self.db)
            .unwrap()
            .unwrap()
            .modify(&self.db, |app| app.contents.default_pot = Some(pot.clone()))
            .map_err(|e| anyhow!(e))
    }

    pub fn get_default_pot(&self, id: AppId) -> anyhow::Result<Option<Pot>> {
        let pot_id = DbApp::get(&AsKey::new(id), &self.db)?
            .map(|app| app.contents.default_pot)
            .ok_or(Error::msg("App not found"))?
            .ok_or(Error::msg("Could not get default pot"))?;
        self.get_pot(pot_id)
    }

    pub fn get_all_apps(&self) -> anyhow::Result<Vec<App>> {
        Ok(DbApp::all(&self.db)
            .query()?
            .iter()
            .map(|app| app.contents.clone().into())
            .collect_vec())
    }

    pub fn get_all_app_ids(&self) -> anyhow::Result<Vec<AppId>> {
        Ok(DbApp::all(&self.db)
            .query()?
            .iter()
            .map(|app| (*app.contents.id).clone())
            .collect_vec())
    }

    pub fn app_has_access(&self, app: AppId, element: ElementId) -> anyhow::Result<bool> {
        if let Some(el) = self.get_element(element) {
            Ok(self
                .get_pot_members(
                    el.pot()
                        .clone()
                        .ok_or(Error::msg("Could not fetch pot members"))?,
                )?
                .contains(&app))
        } else {
            Err(Error::msg("Element not found"))
        }
    }

    pub fn get_app(&self, id: AppId) -> anyhow::Result<Option<App>> {
        DbApp::get(&AsKey::new(id), &self.db)
            .map(|app_option| app_option.map(|app| app.contents.into()))
            .map_err(|e| anyhow!(e))
    }
}

#[cfg(test)]
mod tests {
    use ubisync_lib::types::{AppId, ElementContent, ElementId, Pot, PotId};

    use crate::{api::v0::app::App, state::database::StateDB};

    #[test]
    fn add_get() {
        let db = StateDB::init(None);
        let app_id = AppId::new();
        db.add_app(
            app_id.clone(),
            "app name".to_string(),
            "description".to_string(),
            "app_type".to_string(),
        )
        .unwrap();

        let retrieved_app = db.get_app(app_id.clone()).unwrap();

        match retrieved_app {
            Some(App {
                id,
                app_type,
                name,
                description,
                default_pot,
                last_access: _,
            }) => {
                assert_eq!(
                    (id, app_type, name, description, default_pot),
                    (
                        app_id,
                        "app_type".to_string(),
                        "app name".to_string(),
                        "description".to_string(),
                        None
                    )
                )
            }
            None => assert!(false),
        }
    }

    #[test]
    fn get_default_pot() {
        let db = StateDB::init(None);
        let app_id = AppId::new();
        db.add_app(
            app_id.clone(),
            "app name".to_string(),
            "description".to_string(),
            "app_type".to_string(),
        )
        .unwrap();

        assert_eq!(db.get_default_pot(app_id).unwrap(), None)
    }

    #[test]
    fn set_default_pot() {
        let db = StateDB::init(None);
        let app_id = AppId::new();
        db.add_app(
            app_id.clone(),
            "app name".to_string(),
            "description".to_string(),
            "app_type".to_string(),
        )
        .unwrap();

        let pot = Pot::new(PotId::new(), "app_type".to_string());
        db.add_pot(pot.id.clone(), pot.app_type.clone()).unwrap();
        db.set_default_pot(app_id.clone(), pot.id.clone()).unwrap();

        assert_eq!(db.get_default_pot(app_id).unwrap(), Some(pot))
    }

    #[test]
    fn get_apps() {
        let db = StateDB::init(None);

        assert_eq!(db.get_all_apps().unwrap(), vec![]);

        let (app1, app2) = (AppId::new(), AppId::new());

        db.add_app(
            app1,
            "name1".to_string(),
            "desc1".to_string(),
            "type1".to_string(),
        )
        .unwrap();
        db.add_app(
            app2,
            "name2".to_string(),
            "desc2".to_string(),
            "type2".to_string(),
        )
        .unwrap();

        assert_eq!(db.get_all_apps().unwrap().len(), 2);
    }

    #[test]
    fn get_app_ids() {
        let db = StateDB::init(None);
        assert_eq!(db.get_all_app_ids().unwrap(), vec![]);

        let (app1, app2) = (AppId::new(), AppId::new());

        db.add_app(
            app1.clone(),
            "name1".to_string(),
            "desc1".to_string(),
            "type1".to_string(),
        )
        .unwrap();
        db.add_app(
            app2.clone(),
            "name2".to_string(),
            "desc2".to_string(),
            "type2".to_string(),
        )
        .unwrap();

        assert_eq!(db.get_all_app_ids().unwrap(), vec![app1, app2])
    }

    #[test]
    fn app_access() {
        let db = StateDB::init(None);
        let app_id = AppId::new();
        let pot_id = PotId::new();
        let element_id = ElementId::new();

        db.add_app(
            app_id.clone(),
            "name".to_string(),
            "description".to_string(),
            "app_type".to_string(),
        )
        .unwrap();

        db.add_pot(pot_id.clone(), "app_type".to_string()).unwrap();
        db.add_element(
            element_id.clone(),
            ElementContent::Text("Text".to_string()),
            None,
            false,
            pot_id.clone(),
        )
        .unwrap();

        assert_eq!(
            db.app_has_access(app_id.clone(), element_id.clone())
                .unwrap(),
            false
        );

        db.add_pot_membership(pot_id, app_id.clone()).unwrap();
        assert_eq!(db.app_has_access(app_id, element_id).unwrap(), true);
    }
}
